import calendar
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_query.extension import List
from cpl_translation import TranslatePipe
from discord import Guild, PrivacyLevel
from discord.ext import tasks
from discord.scheduled_event import ScheduledEvent as DiscordEvent

from bot_core.abc.task_abc import TaskABC
from bot_core.logging.task_logger import TaskLogger
from bot_core.service.message_service import MessageService
from bot_data.abc.scheduled_event_repository_abc import ScheduledEventRepositoryABC
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.model.scheduled_event import ScheduledEvent
from bot_data.model.scheduled_event_interval_enum import ScheduledEventIntervalEnum


class ScheduledEventsWatcher(TaskABC):
    def __init__(
        self,
        config: ConfigurationABC,
        logger: TaskLogger,
        bot: DiscordBotServiceABC,
        db: DatabaseContextABC,
        servers: ServerRepositoryABC,
        events: ScheduledEventRepositoryABC,
        message_service: MessageService,
        t: TranslatePipe,
    ):
        TaskABC.__init__(self)

        self._config = config
        self._logger = logger
        self._bot = bot
        self._db = db
        self._servers = servers
        self._events = events
        self._message_service = message_service
        self._t = t

        if not self._is_maintenance():
            self.watch.start()

    def _append_interval(self, interval: ScheduledEventIntervalEnum, ts: datetime) -> datetime:
        now = datetime.now()
        if ts >= now:
            return ts

        if interval == ScheduledEventIntervalEnum.daily:
            ts = ts + timedelta(days=1)

        elif interval == ScheduledEventIntervalEnum.weekly:
            ts = ts + timedelta(weeks=1)

        elif interval == ScheduledEventIntervalEnum.monthly:
            days_in_month = calendar.monthrange(ts.year, ts.month + 1)[1]
            ts = ts + timedelta(days=days_in_month)

        elif interval == ScheduledEventIntervalEnum.yearly:
            ts = ts + timedelta(days=365)

        return ts

    @tasks.loop(hours=24)
    async def watch(self):
        self._logger.info(__name__, "Watching scheduled events")
        try:
            for guild in self._bot.guilds:
                guild: Guild = guild
                server = self._servers.get_server_by_discord_id(guild.id)
                scheduled_events_from_guild = self._events.get_scheduled_events_by_server_id(server.id)
                for scheduled_event in scheduled_events_from_guild:
                    scheduled_event: ScheduledEvent = scheduled_event
                    from_guild = List(DiscordEvent, guild.scheduled_events).where(
                        lambda x: x.name == scheduled_event.name
                        and x.description == scheduled_event.description
                        and x.entity_type == scheduled_event.entity_type
                    )
                    if from_guild.count() != 0:
                        continue

                    kwargs = {"name": scheduled_event.name, "description": scheduled_event.description}

                    if scheduled_event.channel_id is not None:
                        kwargs["channel"] = guild.get_channel(scheduled_event.channel_id)

                    if scheduled_event.start_time is not None:
                        scheduled_event.start_time = self._append_interval(
                            scheduled_event.interval, scheduled_event.start_time
                        )

                        start_time = scheduled_event.start_time.replace(tzinfo=ZoneInfo("Europe/Berlin"))

                        kwargs["start_time"] = start_time

                    if scheduled_event.end_time is not None:
                        scheduled_event.end_time = self._append_interval(
                            scheduled_event.interval, scheduled_event.end_time
                        )
                        end_time = scheduled_event.end_time.replace(tzinfo=ZoneInfo("Europe/Berlin"))
                        kwargs["end_time"] = end_time

                    kwargs["entity_type"] = scheduled_event.entity_type
                    if scheduled_event.location is not None:
                        kwargs["location"] = scheduled_event.location

                    kwargs["privacy_level"] = PrivacyLevel.guild_only

                    try:
                        self._logger.debug(__name__, f"Try to create scheduled event for guild {guild.name}")
                        await guild.create_scheduled_event(**kwargs)
                        self._events.update_scheduled_event(scheduled_event)
                        self._db.save_changes()
                    except Exception as e:
                        self._logger.error(__name__, f"Watching scheduled events failed", e)
        except Exception as e:
            self._logger.error(__name__, f"Watching scheduled events failed", e)

    @watch.before_loop
    async def wait(self):
        await self._wait_until_ready()
